
import logging

# import pytest
# from contextlib import nullcontext as does_not_raise

# # (assuming check_user function is defined above)

# @pytest.mark.parametrize(
#     "name, expected_exception, expected_result",
#     [
#         ("Yana", does_not_raise(), "Username is properly set"),
#         ("Bob", pytest.raises(ValueError, match="Invalid User"), None),
#     ],
# )
# def test_check_user_parametrized(name, expected_exception, expected_result):
#     with expected_exception:
#         result = check_user(name)
#         if expected_result is not None:
#             assert result == expected_result




import logging
from typing import Annotated, Optional
import inspect
import asyncio
from pprint import pprint
from pathlib import Path
from zen_rule import ZenRule, udf, FuncArg, FuncRet

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


@udf(
    comments="test udf foo",
    args_info=[
        FuncArg(arg_name="a", arg_type="string", defaults="", comments="var a"),
        FuncArg(arg_name="b", arg_type="string", defaults="", comments="var b"),
        FuncArg(arg_name="c", arg_type="string", defaults="", comments="var c"),
    ],
    return_info=FuncRet(field_type="string", examples="fccdjny", comments="返回值示例, 字段解释")     
)
def foo(*args, **kwargs):
    logger.info(f"{inspect.stack()[0][3]} args:{args}")
    logger.info(f"{inspect.stack()[0][3]} kwargs:{kwargs}")
    return "foo value"


@udf()
def bar(c: int, b:Annotated[str, "Annotated tips for arg b"],
        a: int, z: int = "fccdjny", *args, **kwargs)-> int:
    """
    Docstring for bar short description only one line
    Docstring for bar long description multiple line, end by first :param line
    1
    2

    3
    
    :param a: arg a
    :type a: int
    :param b: arg b
    :param c: arg c
    :type c: int
    :param args: param args
    :param d: Description
    :type d: int
    :param kwargs: param kwargs
    """
    pass


if __name__ == "__main__":
    from pprint import pprint
    from zen_rule import ZenRule
    # from zen_rule import udf_manager
    # pprint(udf_manager.udf_function_schema_tools(), sort_dicts=False)
    # print("-------------------")
    # pprint(udf_manager.get_udf_info(), sort_dicts=False)
    pprint(ZenRule.udf_function_schema_tools(), sort_dicts=False)



import logging
import inspect
from .udf import udf, FuncArg
logger = logging.getLogger(__name__)


@udf(
    comments="default in out functions, Keep parameters unchanged output for debug",
    namespace="default",
    args_info=[
        FuncArg(arg_name="b", arg_type="string", defaults="", comments="var b"),
        FuncArg(arg_name="a", arg_type="string", defaults="", comments="var a"),
        FuncArg(arg_name="c", arg_type="string", defaults="", comments="var c"),
    ],  # todo: 以后考虑从入参定义中提取入参类型和名字, 从 Docstring 中提取参数和返回值的注释.
    return_info=None     
)
def inout(*args, **kwargs) -> str:
    """
    Docstring for inout
    自定义函数测试, 返回值返回入参, 用于调试.
    
    :param args: Description
    :param kwargs: Description
        kwargs 包含 zen-rule 传入的节点入参和元信息.
        其中 _node_input 表示此节点的所有入参.
    """
    logger.info(f"{inspect.stack()[0][3]} args:{args}")
    logger.info(f"{inspect.stack()[0][3]} kwargs:{kwargs}")
    return kwargs.get("_node_input_", {})


@udf(
    comments="default in out functions, Keep parameters unchanged output for debug",
    namespace="default",
    args_info=[],  # todo: 以后考虑从入参定义中提取入参类型和名字, 从 Docstring 中提取参数和返回值的注释.
    return_info=None     
)
def func_without_args(*args, **kwargs) -> str:
    """
    Docstring for func_without_args
    无参数函数, 用于自定义函数测试
    
    :param args: Description
    :param kwargs: Description
        kwargs 包含 zen-rule 传入的节点入参和元信息.
        其中 _node_input 表示此节点的所有入参.
    """
    logger.info(f"{inspect.stack()[0][3]} args:{args}")
    logger.info(f"{inspect.stack()[0][3]} kwargs:{kwargs}")
    return kwargs.get("_node_input_", {})


import logging
import inspect
from itertools import groupby
from pprint import pformat
from typing import List, Dict, Any
from inspect import signature, Parameter, iscoroutinefunction
from pydantic import create_model, Field, TypeAdapter
from docstring_parser import parse

logger = logging.getLogger(__name__)

py_json_type_maps = {
    None: 'null',
    bool: 'boolean',
    str: 'string',
    dict: 'object',
    list: 'array',
    int: 'integer',
    float: 'number'
}
json_py_type_maps = {v:k for k,v in py_json_type_maps.items()}
py_json_type_maps[tuple] = "array"


def namespace_split(target,namespace=None) -> str:
    if namespace:
        return str(namespace)
    try:
        name_list = str(target).split(".")
        new_name_list = []
        for item in name_list:
            if "functions" == item:
                new_name_list = []
            else:
                new_name_list.append(item)
        return ".".join(new_name_list)
    except Exception as e:
        logger.error(e, exc_info=True)
        return ""


# Define classes for managing function arguments and return values
class FuncArg:
    def __init__(self, arg_name: str, arg_type: str, defaults: Any, comments: str):
        self.arg_name = arg_name
        self.arg_type = arg_type
        self.defaults = defaults
        self.comments = comments

    def to_dict(self):
        return {
            "arg_name": self.arg_name,
            "arg_type": self.arg_type,
            "defaults": self.defaults,
            "comments": self.comments,
        }

class FuncRet:
    def __init__(self, field_type: str, examples: Any, comments: str):
        self.field_type = field_type
        self.examples = examples
        self.comments = comments

    def to_dict(self):
        return {
            "field_type": self.field_type,
            "examples": self.examples,
            "comments": self.comments,
        }

def function_return_schema(f):
    func_sig = inspect.signature(f)
    if func_sig.return_annotation == Parameter.empty:
        # 如果返回值未定义那么返回空属性的的 return_model
        return_schema = {
            'properties': {},
            'title': 'non return annotation',
            'type': 'object',
        }
    elif func_sig.return_annotation in {str, int, float, bool}:
        return_schema = {
            'properties': {},
            'title': '',
            'type': py_json_type_maps.get(func_sig.return_annotation),
        }
    elif func_sig.return_annotation in {list, tuple}:
        ## from list, tuple generate json schema
        return_schema = {
            'properties': {},
            'title': '',
            'type': py_json_type_maps.get(func_sig.return_annotation),
        }
    elif func_sig.return_annotation in {dict}:
        ## from dict generate json schema
        return_model = create_model(f.__name__ + "_return", **{
            # field: (field_type, default_value)
        })
        return_schema = return_model.model_json_schema()
    else:
        ## func_sig.return_annotation 是一个实例类型，可以转化为字典
        ## 把基于实例的字典的键值对放入第二个参数, 生成函数返回值的json schema描述.
        return_model = create_model(f.__name__ + "_return", **{
            # field: (field_type, default_value)
        })
        return_schema = return_model.model_json_schema()
    return return_schema


def function_schema(f):
    """
        todo: function_schema 增加关于返回值schema的字段定义.
        todo: function_schema 按照 namespace 分组.
    """
    ### def bar(a: int, b:Annotated[str, "The city to get the weather for"]):
    ###     pass
    func_sig = inspect.signature(f)
    func_parameters = {}
    parsed_doc = parse(f.__doc__)
    param_docs = {param.arg_name:param for param in parsed_doc.params}
    for n, param in func_sig.parameters.items():
        if n == "args" or n == "kwargs":
            continue
        # Retrieve function argument Annotated tips
        _para_annotation_doc = getattr(param.annotation, "__metadata__", "")
        para_annotation_doc = _para_annotation_doc[0] if _para_annotation_doc else ""
        t = None if param.annotation==Parameter.empty else param.annotation
        default_value = Ellipsis if param.default==Parameter.empty else param.default
        # comment from doc string or field annotation, field annotation first
        _param_func_doc = param_docs.get(n)
        param_func_doc = _param_func_doc.description if _param_func_doc else ""
        field = Field(default_value, description=para_annotation_doc or param_func_doc)
        func_parameters[n] = (t, field)

    parameter_model = create_model(f.__name__, **func_parameters)
    return_schema = function_return_schema(f)
    # function json schema
    func_docs = [parsed_doc.short_description, parsed_doc.long_description]
    func_description = "\n".join([i for i in func_docs if i])  # 取出非Node的 doc string
    schema = {
        'name': f.__name__,   # name是 openai function calling schema 的关键字
        'title': f.__name__,  # title 是 json schema 的关键字
        'type': 'function',
        'description': func_description,
        'parameters': parameter_model.model_json_schema(),
        # 'return': return_schema
    }
    return schema


# UDF Manager class to manage the registered UDFs
class UDFManager:
    # 之后再考虑是否支持 zen expression 的混用了.
    def __init__(self):
        self.functions = {}
        self.funcs = {}  # function schema dict
        self.param_annotation = {
            str: "string",
            dict: "object",  # json
            list: "array",
            tuple: "array",
            int: "number",
            float: "number",
        }
        self.default_type = "Any"

    def register_function(self, func, comments:str =None, args_info: FuncArg =None, return_info: FuncRet =None, namespace:str = None):
        sig = signature(func)
        arguments = []

        # Use provided args_info or derive from function signature
        if args_info:
            for info in args_info:
                arguments.append(info)
        else:
            # logger.debug(f"sig.parameters: {sig.parameters}")
            # logger.debug(f"sig.parameters items(): {pformat(sig.parameters.items())}")
            for name, param in sig.parameters.items():
                arg_type = self.param_annotation.get(param.annotation, self.default_type)
                defaults = '' if param.default == Parameter.empty else param.default
                comments = f'{name}:{arg_type}'
                arguments.append(FuncArg(name, arg_type, defaults, comments))

            # extra_arguments = [(arg.arg_name, arg.arg_type) for arg in arguments if arg.arg_name not in {"args", "kwargs"}]
            # if extra_arguments:
            #     raise Exception("udf function only allow to contains *args, **kwargs, please use @udf args_info to describe arg names")            
        # logger.debug(f"return_info:{return_info}")
        return_values = return_info.to_dict() if return_info else {}
        # logger.debug(f"return_values:{return_values}")
        self.functions[func.__name__] = {
            "func": func,
            "name": func.__name__,
            "arglength": len(arguments),
            "arguments": [arg.to_dict() for arg in arguments],
            "return_values": return_values,
            "comments": comments or func.__doc__,
            "namespace": namespace_split(func.__module__,namespace=namespace),
        }
        ### todo: 在 1.0 版本之前移除此兼容逻辑.
        ## 目前函数参数的定义都是 args 和 kwargs.
        ## 实际参数定义在udf装饰器中的args_info列表中的 FuncArg 实例上.
        ## 需要将这部分的参数定义也包含在 function json schema 的 parameters 参数中.
        func_schema = function_schema(func)
        decorated_args = [i for i in self.functions[func.__name__]["arguments"]
                          if i.get("arg_name") not in {"args", "kwargs"}]
        if decorated_args:
            d = {i.get("arg_name"): 
            (
                json_py_type_maps.get(i.get("arg_type"), None),
                Field(i.get("defaults"), description=i.get("comments"))
            )
            for i in decorated_args}
            m = create_model(func.__name__, **d)
            func_schema['parameters'] = m.model_json_schema()
        func_schema['namespace'] = namespace_split(func.__module__, namespace=namespace)
        func_schema['kind'] = func_schema['namespace'].split('.')[-1]
        self.funcs[func.__name__] = func_schema

    def get_udf_info(self) -> List[Dict[str, Any]]:
        return [
            {
                "name": data["name"],
                "arglength": data["arglength"],
                "arguments": data["arguments"],
                "return_values": data["return_values"],
                "comments": data["comments"],
                "namespace": data["namespace"],
                "kind": data["namespace"].split(".")[-1]
            }
            for data in self.functions.values()
        ]

    def udf_info(self, name) -> List[Dict[str, Any]]:
        return self.functions.get(name)

    def udf_function_schema(self, name: str) -> Dict[str, Any]:
        return self.funcs.get(name)

    def udf_function_schema_tools(self) -> Dict[str, Any]:
        d = []
        ## groupby list must sorted
        func_tools = [data for data in self.funcs.values()]
        group_key_func = lambda x: x.get("namespace", "default")
        for key, group in groupby(sorted(func_tools, key=group_key_func), key=group_key_func):
            # simulate openai function calling schema
            namespace_funcs = {
                "type": "namespace",
                "name": key,
                "description": "",  # 以后从模块的顶层文档中提取或者增加显示定义.
                "tools": list(group)
            }
            d.append(namespace_funcs)
        return d

    def func_bind_params(self, name: str, args: list):
        """
            将参数值和参数名字组装为字典.
            f["parameters"]["properties"] 是如下结构
            {
                'b': {'default': '', 'description': 'var b', 'title': 'B', 'type': 'string'},
                'a': {'default': '', 'description': 'var a', 'title': 'A', 'type': 'string'},
                'c': {'default': '', 'description': 'var c', 'title': 'C', 'type': 'string'}
            }
        """
        f = self.udf_function_schema(name)
        param_with_values = {name: v for (name, _), v in zip(f["parameters"]["properties"].items(), args)}
        return param_with_values

    async def __call__(self, udf_name: str, *args, **kwargs) -> Any:
        """
            v2 spec
            执行器要函数解析放到一起, 在function_call_parse中实现解析器和执行器,
            然后执行器调用真正的udf函数管理和执行器.
        """
        if udf_name in self.functions:
      
            func = self.functions[udf_name]["func"]
            if iscoroutinefunction(func):
                return await func(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        else:
            raise ValueError(f"Function '{udf_name}' is not registered in UDFManager")

    async def call_udf(self, udf_name: str, *args, **kwargs) -> Any:

        if udf_name in self.functions:
      
            func = self.functions[udf_name]["func"]
            if iscoroutinefunction(func):
                return await func(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        else:
            raise ValueError(f"Function '{udf_name}' is not registered in UDFManager")

# Instantiate UDF Manager
udf_manager = UDFManager()

# Decorator to register UDFs with optional custom metadata
def udf(comments=None, args_info=None, return_info=None,namespace=None):
    def decorator(func):
        udf_manager.register_function(func, comments, args_info, return_info,namespace)
        return func
    return decorator



import zen
import pytest
import warnings
warnings.warn("This test is deprecated and will be removed in the future.", DeprecationWarning)
## 未来再合并到 zen 表达式的测试用例中
def test_zen_exprs():
    """
        @Deprecated: 此测试用例是想把 zen 表达式用 python 来解析.
        但是 zen 表达式没有明确的语法格式定义. 而且未来变动比较大.
        需要case by case去支持. 还不如直接在 zen-expression 的 rust 程序中去支持拓展.
    """
    assert zen.evaluate_expression("{a:1}", {"a":3, "b":"fccdjny", "c": {'name': 'Alice', 'age': 30}})
    assert zen.evaluate_expression("{'a':1}", {"a":3, "b":"fccdjny", "c": {'name': 'Alice', 'age': 30}})
    assert zen.evaluate_expression('{"a":1}', {"a":3, "b":"fccdjny", "c": {'name': 'Alice', 'age': 30}})

    ## 注意, zen context 的键值对中的键是字符串，这个字符串可以用单引号，双引号，也可以不用引号(像一个变量一样)
    assert zen.evaluate_expression('{ firstName: "John", lastName: last }',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"}) == {'lastName': 'fccdjny', 'firstName': 'John'}
    assert zen.evaluate_expression("{ 'firstName': 'John', lastName: last }",  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"}) == {'lastName': 'fccdjny', 'firstName': 'John'}
    assert zen.evaluate_expression('{ "firstName": "John", lastName: last }',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"}) == {'lastName': 'fccdjny', 'firstName': 'John'}
    assert set(zen.evaluate_expression('values({ firstName: "John", lastName: last })',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"})) == set(['fccdjny', 'John'])
    assert set(zen.evaluate_expression('keys({ firstName: "John", lastName: last })',  {"firstName":"fff", "lastName": "lll", "last":"fccdjny"})) == set(['lastName', 'firstName'])
    ## 测试 zen 表达式 变量数组 下标访问   支持
    assert zen.evaluate_expression('a[1]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == 2
    ## 测试 zen 表达式 变量数组 切片   支持
    assert zen.evaluate_expression('a[1:]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == [2,3,4,5]
    assert zen.evaluate_expression('a[1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == [2,3]

    ## 测试 zen 表达式 变量字符串 下标访问  不支持  结果为None
    assert zen.evaluate_expression('b[1]', {"a": [1,2,3,4,5],"b":"fccdjny"}) is None

    ## 测试 zen 表达式 变量字符串 切片  支持
    assert zen.evaluate_expression('b[0:0]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == "f"
    assert zen.evaluate_expression('b[1:]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == "ccdjny"
    assert zen.evaluate_expression('b[1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == "cc"

    ## 测试 zen 表达式 常量数组 切片    支持
    assert zen.evaluate_expression('[1,2,3,4,5][1]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == 2
    assert zen.evaluate_expression('[1,2,3,4,5][1:]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == [2,3,4,5]
    assert zen.evaluate_expression('[1,2,3,4,5][1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == [2,3]

    ## 测试 zen 表达式 常量字符串 切片  不支持
    with pytest.raises(RuntimeError, match="parserError"):
        assert zen.evaluate_expression('"fccdjny"[1]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == 'c' 
    with pytest.raises(RuntimeError, match="parserError"):
        assert zen.evaluate_expression('"fccdjny"[1:]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == 'ccdjny' 
    with pytest.raises(RuntimeError, match="parserError"):
        assert zen.evaluate_expression('"fccdjny"[1:2]', {"a": [1,2,3,4,5],"b":"fccdjny"}) == 'cc'

    ## 测试 zen 表达式 变量字符串 切片 下标非整数  支持
    assert zen.evaluate_expression('a[3: 3+3]', {"a": "fccdjny"}) == "djny"
    assert zen.evaluate_expression('a[3: 3+1.3]', {"a": "fccdjny"}) == "dj"

    ## 测试 zen 表达式 字符串插值
    assert zen.evaluate_expression('`User ${user.name} is ${user.age} years old`', {"user":{"name":"John","age":30}}) == "User John is 30 years old"
    assert zen.evaluate_expression("'User ${user.name} is ${user.age} years old'", {"user":{"name":"John","age":30}}) == 'User ${user.name} is ${user.age} years old'

    ## 测试 zen 表达式 嵌套访问
    assert zen.evaluate_expression('user.contacts[i].phone.p[0]',
                                   {"i":1,
                                    "user":{"contacts":
                                            [
                                                {"phone":"123-456-7890"},
                                                {"phone":{"p":["456-123-7890"]}}
                                            ]
                                            }
                                    }) == '456-123-7890'
    assert zen.evaluate_expression('user.contacts [i].phone.p[0]', 
                                   {"i":1,
                                    "user":{"contacts":
                                            [
                                                {"phone":"123-456-7890"}, 
                                                {"phone":{"p":["456-123-7890"]}}
                                            ]
                                            }
                                    }) == '456-123-7890'

    ## 三元表达式和非空判断
    assert zen.evaluate_expression('a? 1 :2 + 3', {"a": True}) == 1
    assert zen.evaluate_expression('(a? 1 : 2) + 3', {"a": True}) == 4
    assert zen.evaluate_expression('a? b ?? "bbb" :2 + 3', {"a": True}) == "bbb"



import timeit
import argparse
import zen
from zen import ZenEngine, ZenDecision
graph_content = """
{
  "contentType": "application/vnd.gorules.decision",
  "nodes": [
    {
      "type": "inputNode",
      "content": {
        "schema": ""
      },
      "id": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "name": "request",
      "position": {
        "x": 245,
        "y": 355
      }
    },
    {
      "type": "outputNode",
      "content": {
        "schema": ""
      },
      "id": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a",
      "name": "response",
      "position": {
        "x": 1120,
        "y": 365
      }
    },
    {
      "type": "expressionNode",
      "content": {
        "expressions": [
          {
            "id": "8bf2c094-3115-4608-9416-1c854c68cc89",
            "key": "result",
            "value": "num * 2"
          }
        ],
        "passThrough": false,
        "inputField": null,
        "outputPath": null,
        "executionMode": "single"
      },
      "id": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "name": "expression1",
      "position": {
        "x": 715,
        "y": 400
      }
    }
  ],
  "edges": [
    {
      "id": "84fa9a9a-e90a-4322-be6f-94d3efd84edf",
      "sourceId": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "type": "edge",
      "targetId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e"
    },
    {
      "id": "0cd7ec7d-05d9-4497-abc7-d51b2202ddb4",
      "sourceId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "type": "edge",
      "targetId": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a"
    }
  ]
}
"""


def init_zen():
    engine = zen.ZenEngine()
    decision = engine.create_decision(graph_content)
    return graph_content, engine, decision


con, engine, decision = init_zen()


def demo1(con, engine, decision):
    decision = engine.create_decision(con)
    input = {
      "num": 10,
    }
    result = decision.evaluate(input, {"trace": True})
    return result


def demo2(con, engine, decision):
    input = {
      "num": 10,
    }
    result = decision.evaluate(input, {"trace": True})
    return result


def demo3(con, engine, decision):
    input = {
      "num": 10,
    }
    result = decision.evaluate(input, {"trace": False})
    return result


# 1. Create the ArgumentParser object
parser = argparse.ArgumentParser(description="test different zen engine usage demo performace")

# 2. Add arguments
parser.add_argument("name", help="name of test function.", choices=['demo1', 'demo2', 'demo3', 'bench'])
# parser.add_argument("bench", choices=['d', 's', 'b'], help="Include a greeting.")
# parser.print_help()

# 3. Parse the arguments
args = parser.parse_args()

# 4. Access and use the arguments
from pprint import pprint
print(args)
if args.name in globals():
  print(globals()[args.name](con, engine, decision))
else:
  if args.name == 'bench':
    number = 10000
    r1 = timeit.timeit('demo1(con, engine, decision)', number=number, globals=globals())
    print(f"demo1: {r1} s/{number}ops")
    r2 = timeit.timeit('demo2(con, engine, decision)', number=number, globals=globals())
    print(f"demo2: {r2} s/{number}ops")
    r3 = timeit.timeit('demo3(con, engine, decision)', number=number, globals=globals())
    print(f"demo3: {r3} s/{number}ops")
  else:
    print(f"{args.name} function not found")






import zen
from zen import ZenEngine, ZenDecision
graph_content = """
{
  "contentType": "application/vnd.gorules.decision",
  "nodes": [
    {
      "type": "inputNode",
      "content": {
        "schema": ""
      },
      "id": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "name": "request",
      "position": {
        "x": 245,
        "y": 355
      }
    },
    {
      "type": "outputNode",
      "content": {
        "schema": ""
      },
      "id": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a",
      "name": "response",
      "position": {
        "x": 1120,
        "y": 365
      }
    },
    {
      "type": "expressionNode",
      "content": {
        "expressions": [
          {
            "id": "8bf2c094-3115-4608-9416-1c854c68cc89",
            "key": "result",
            "value": "num * 2"
          }
        ],
        "passThrough": false,
        "inputField": null,
        "outputPath": null,
        "executionMode": "single"
      },
      "id": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "name": "expression1",
      "position": {
        "x": 715,
        "y": 400
      }
    }
  ],
  "edges": [
    {
      "id": "84fa9a9a-e90a-4322-be6f-94d3efd84edf",
      "sourceId": "d4c5fdef-96c0-4c5e-8e7b-e1e4f9797251",
      "type": "edge",
      "targetId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e"
    },
    {
      "id": "0cd7ec7d-05d9-4497-abc7-d51b2202ddb4",
      "sourceId": "45bbd693-493b-415d-8ad6-0d57c9a40d3e",
      "type": "edge",
      "targetId": "a2cb6433-140b-4d2c-89f2-0e4f3140dd1a"
    }
  ]
}
"""
def init_zen():
  engine = zen.ZenEngine()
  decision = engine.create_decision(graph_content)
#   zen_decison_content = ZenDecisionContent(con)
  return graph_content,engine, decision


con, engine, decision = init_zen()


def zenengine_demo1(con, engine, decision):
  decision = engine.create_decision(con)
  input = {
    "num": 10,
  }
  result = decision.evaluate(input, {"trace": True})

def zenengine_demo2(con, engine, decision):
  input = {
    "num": 10,
  }
  result = decision.evaluate(input, {"trace": True})

def zenengine_demo3(con, engine, decision):
  input = {
    "num": 10,
  }
  result = decision.evaluate(input, {"trace": False})


# def zenengine_demo1_1(con, engine, decision):
#   decision = engine.create_decision(zen_decison_content)
#   input = {
#     "num": 10,
#   }
#   result = decision.evaluate(input, {"trace": True})

# def zenengine_demo1_2(con, engine, decision):
#   decision = engine.create_decision(zen_decison_content)
#   input = {
#     "num": 10,
#   }
#   result = decision.evaluate(input)


# %timeit zenengine_demo1(con, engine, decision)
# %timeit zenengine_demo2(con, engine, decision)
# %timeit zenengine_demo3(con, engine, decision)


# %timeit zenengine_demo1_1(con, engine, decision)
# %timeit zenengine_demo1_2(con, engine, decision)

# 290000/600
# 1000000/290


from typing import Any, Optional, TypedDict, Literal, Awaitable, Union
import logging
import json
from pathlib import Path
from pprint import pprint, pformat
import asyncio
import inspect
import contextvars

import zen
from zen import ZenDecision
from utils import httpsession
# from zen import EvaluateResponse  # cannot import
zen_exprs_eval = lambda x, input: zen.evaluate_expression(x, input)
logger = logging.getLogger(__name__)
logging.basicConfig(level="NOTSET")


class EvaluateResponse(TypedDict):
    performance: str
    result: dict
    trace: dict


# async def custom_async_handler(request):
#     print("custom_async_handler custom_async_handler custom_async_handler")
#     p1 = request.get_field("prop1")
#     await asyncio.sleep(0.1)
#     return {
#         "output": {"sum": p1}
#     }


async def custom_async_handler(request):
    """
        示例自定义节点处理函数.
        todo: 是否需要包装一层异常捕获的逻辑.
    """
    try:
        # p1 = request.get_field("prop1")  # 没有prop1属性会报错.
        # await asyncio.sleep(0.1)
        ctx = contextvars.copy_context()
        print("custom_async_handler contextvars:")
        pprint(dict(ctx.items()))
        print("custom_async_handler -> httpsession.get():", httpsession.get())
        print("custom_async_handler -> httpsession.set('custom_async_handler'):")
        httpsession.set("custom_async_handler")
        print("custom_async_handler -> httpsession.get():", httpsession.get())
        logger.debug(f"request:{request}")
        logger.debug(f"request attrs:{dir(request)}")
        result = zen.evaluate_expression('rand(100)', request.input)
        logger.debug(f"return value:{result}")
        await asyncio.sleep(0)
        return {
            "output": {"sum": 112}
        }
    except Exception as e:
        import traceback
        traceback.print_exc()
        raise e


def read_json_graph(filename):
    with open(filename, "r", encoding="utf8") as f:
        logger.warning(f"graph json: %s", filename)
        content =  f.read()
    return content



async def init_zen_engine():
    # httpsession.set("fccdjny")  # 先设置 contextvars 再创建包含 custom_handler 的zen engine 实例, 自定义函数可以 contextvars get.
    # print("init_zen_engine -> httpsession.get():", httpsession.get())
    ## options 中 customHandler 为 async_function, 必须放在异步函数中.
    ## 否则会报错 sys:1: RuntimeWarning: coroutine 'custom_async_handler' was never awaited
    ## custom_async_handler 中的 contextvars 必须在 zen.ZenEngine 初始化之前设置.
    ## 否则 custom_async_handler 中无法获取相关的contextvars.
    options = {"customHandler": custom_async_handler}
    engine = zen.ZenEngine(options) 
    basedir = Path(__file__).parent
    filename = basedir / "graph" / "custom_v0.json"
    content = read_json_graph(filename=filename)
    httpsession.set("fccdjny2")  # zen.ZenEngine 初始后设置 contextvars, custom_async_handler 中无法获取.
    zendecision = engine.create_decision(content)
    result = await zendecision.async_evaluate({"input": 7, "myvar": 15})
    print("init_zen_engine -> httpsession.get():", httpsession.get())
    return result


result = asyncio.run(init_zen_engine())
print("result:", result)


# if __name__ == "__main__":
#     pass


def sort_seven():
    """Sort a list of seven items"""
    for _ in range(10_000):
        sorted([3,2,4,5,1,5,3])

def sort_three():
    """Sort a list of three items"""
    for _ in range(10_000):
        sorted([3,2,4])

__benchmarks__ = [
    (sort_seven, sort_three, "Sorting 3 items instead of 7")
]